#include <stdlib.h>
#include <dos.h>
#include <conio.h>
#include "ns_dpmi.h"
#include "ns_task.h"
#include "ns_cards.h"
#include "ns_user.h"
#include "ns_pwm.h"
#include "ns_muldf.h"
#include "ns_inter.h"

#include "m_misc.h"
#include "options.h"
#include "doomstat.h"

static int PCSpeaker_PWM_Installed = 0;

static char *PCSpeaker_PWM_BufferStart;
static char *PCSpeaker_PWM_CurrentBuffer;
static int PCSpeaker_PWM_BufferNum = 0;
static int PCSpeaker_PWM_NumBuffers = 0;
static int PCSpeaker_PWM_TransferLength = 0;
static int PCSpeaker_PWM_CurrentLength = 0;

static char *PCSpeaker_PWM_SoundPtr;
volatile int PCSpeaker_PWM_SoundPlaying;

static task *PCSpeaker_PWM_Timer;

void (*PCSpeaker_PWM_CallBack)(void);

/*---------------------------------------------------------------------
   Function: PCSpeaker_PWM_ServiceInterrupt

   Handles interrupt generated by sound card at the end of a voice
   transfer.  Calls the user supplied callback function.
---------------------------------------------------------------------*/

static void PCSpeaker_PWM_ServiceInterrupt(task *Task)
{
    unsigned char value = (unsigned char) *PCSpeaker_PWM_SoundPtr;

    OutByte43h(0xB0);
    OutByte42h(value >> 1);
    OutByte42h(0);

    PCSpeaker_PWM_SoundPtr++;

    PCSpeaker_PWM_CurrentLength--;
    if (PCSpeaker_PWM_CurrentLength == 0)
    {
        // Keep track of current buffer
        PCSpeaker_PWM_CurrentBuffer += PCSpeaker_PWM_TransferLength;
        PCSpeaker_PWM_BufferNum++;
        if (PCSpeaker_PWM_BufferNum >= PCSpeaker_PWM_NumBuffers)
        {
            PCSpeaker_PWM_BufferNum = 0;
            PCSpeaker_PWM_CurrentBuffer = PCSpeaker_PWM_BufferStart;
        }

        PCSpeaker_PWM_CurrentLength = PCSpeaker_PWM_TransferLength;
        PCSpeaker_PWM_SoundPtr = PCSpeaker_PWM_CurrentBuffer;

        // Call the caller's callback function
        if (PCSpeaker_PWM_CallBack != NULL)
        {
            MV_ServiceVoc();
        }
    }
}

/*---------------------------------------------------------------------
   Function: PCSpeaker_PWM_StopPlayback

   Ends the transfer of digitized sound to the Sound Source.
---------------------------------------------------------------------*/

void PCSpeaker_PWM_StopPlayback(void)
{
    if (PCSpeaker_PWM_SoundPlaying)
    {
        TS_Terminate(PCSpeaker_PWM_Timer);
        PCSpeaker_PWM_SoundPlaying = 0;
        PCSpeaker_PWM_BufferStart = NULL;
    }

    // Turn off
    OutByte61h(InByte61h() & 0xFC);
}

/*---------------------------------------------------------------------
   Function: PCSpeaker_PWM_BeginBufferedPlayback

   Begins multibuffered playback of digitized sound on the Sound Source.
---------------------------------------------------------------------*/

int PCSpeaker_PWM_BeginBufferedPlayback(
    char *BufferStart,
    int BufferSize,
    int NumDivisions,
    void (*CallBackFunc)(void))
{
    if (PCSpeaker_PWM_SoundPlaying)
    {
        PCSpeaker_PWM_StopPlayback();
    }

    PCSpeaker_PWM_CallBack = CallBackFunc;

    PCSpeaker_PWM_BufferStart = BufferStart;
    PCSpeaker_PWM_CurrentBuffer = BufferStart;
    PCSpeaker_PWM_SoundPtr = BufferStart;
    // VITI95: OPTIMIZE
    PCSpeaker_PWM_TransferLength = BufferSize / NumDivisions;
    PCSpeaker_PWM_CurrentLength = PCSpeaker_PWM_TransferLength;
    PCSpeaker_PWM_BufferNum = 0;
    PCSpeaker_PWM_NumBuffers = NumDivisions;

    PCSpeaker_PWM_SoundPlaying = 1;

    PCSpeaker_PWM_Timer = TS_ScheduleTask(PCSpeaker_PWM_ServiceInterrupt, PCSpeaker_PWM_SampleRate, 1, NULL);
    TS_Dispatch();

    return (PCSpeaker_PWM_Ok);
}

/*---------------------------------------------------------------------
   Function: PCSpeaker_PWM_Init

   Initializes the Sound Source prepares the module to play digitized
   sounds.
---------------------------------------------------------------------*/

int PCSpeaker_PWM_Init(int soundcard)
{
    if (PCSpeaker_PWM_Installed)
    {
        PCSpeaker_PWM_Shutdown();
    }

    OutByte61h(InByte61h() | 0x3);
    OutByte43h(0xB0);
    OutByte42h(0);
    OutByte42h(0);

    PCSpeaker_PWM_SoundPlaying = 0;

    PCSpeaker_PWM_CallBack = NULL;

    PCSpeaker_PWM_BufferStart = NULL;

    PCSpeaker_PWM_Installed = 1;

    return (PCSpeaker_PWM_Ok);
}

/*---------------------------------------------------------------------
   Function: PCSpeaker_PWM_Shutdown

   Ends transfer of sound data to the Sound Source.
---------------------------------------------------------------------*/

void PCSpeaker_PWM_Shutdown(void)
{
    PCSpeaker_PWM_StopPlayback();

    PCSpeaker_PWM_SoundPlaying = 0;

    PCSpeaker_PWM_BufferStart = NULL;

    PCSpeaker_PWM_CallBack = NULL;

    PCSpeaker_PWM_Installed = 0;
}
